@@ -25,7 +25,6 @@
 #include <stdlib.h>
 #include <stdarg.h>
 #include <string.h>
-#include <stropts.h>
 #include <dirent.h>
 #include <fcntl.h>
 #include <termios.h>
@@ -228,9 +227,9 @@ static void exit_handler (void) attribute((noinline));
  */
 int main(int argc, char *argv[])
 {
-    volatile char *arg0 = argv[0];
     char ptsname[NAME_MAX+1];
     const char *tty, *stt;
+    volatile char *arg0;
     struct console *c;
     struct termios o;
     struct winsize w;
@@ -248,6 +247,9 @@ int main(int argc, char *argv[])
     if (kill (1, SIGRTMIN+20) < 0)
 	warn("could not tell system to show its status");
 
+    arg0 = (volatile char*)argv[0];
+    remember_arg0(arg0);
+
     while ((arg = getopt(argc, argv, "f")) != -1) {
 	switch (arg) {
 	case 'f':
@@ -262,6 +264,7 @@ int main(int argc, char *argv[])
     argc -= optind;
 
     myname = program_invocation_short_name;
+    getconsoles(&cons, 1);
 
     close(0);
     close(1);
@@ -270,6 +273,7 @@ int main(int argc, char *argv[])
 	error("Can not open system console %s", console);
 
     if (fd > 0) {
+	(void)ioctl(fd, TIOCNXCL);	/* Avoid EBUSY */
 	dup2(fd, 0);
 	close(fd);
     }
@@ -280,8 +284,6 @@ int main(int argc, char *argv[])
 
     (void)ioctl(0, TIOCCONS, NULL);  /* Undo any current map if any */
 
-    getconsoles(&cons, 1);
-
     list_for_each_entry(c, &cons->node, node) {
 	speed_t ospeed;
 	speed_t ispeed;
@@ -337,6 +339,10 @@ int main(int argc, char *argv[])
     if (ioctl(pts, TIOCCONS, NULL) < 0)
 	error("can not set console device to %s", ptsname);
 
+    /*
+     * Reconnecting to stdin aka ptm to 0 is done after we fork away
+     */
+
     dup2(pts,  1);
     dup2(pts,  2);	/* Now we are blind upto safeIO() loop */
     if (pts > 2)
@@ -352,10 +358,10 @@ int main(int argc, char *argv[])
     set_signal(SIGQUIT, &saved_sigquit, sighandle);
     set_signal(SIGTERM, &saved_sigterm, sighandle);
     set_signal(SIGSYS,  &saved_sigsys,  sigsys);
-    (void)siginterrupt(SIGINT,  0);
-    (void)siginterrupt(SIGQUIT, 0);
-    (void)siginterrupt(SIGTERM, 0);
-    (void)siginterrupt(SIGSYS,  0);
+    (void)restart_sig(SIGINT,  1);
+    (void)restart_sig(SIGQUIT, 1);
+    (void)restart_sig(SIGTERM, 1);
+    (void)restart_sig(SIGSYS,  1);
 
     list_for_each_entry(c, &cons->node, node) {
 	struct termios oldtio;
@@ -363,7 +369,7 @@ int main(int argc, char *argv[])
 	speed_t ispeed;
 	int flags;
 
-	(void)ioctl(fd, TIOCNXCL);	/* Avoid EBUSY */
+	(void)ioctl(c->fd, TIOCNXCL);	/* Avoid EBUSY */
 
 #ifdef _PC_MAX_CANON
 	if ((c->max_canon = (ssize_t)fpathconf(c->fd, _PC_MAX_CANON)) <= 0)
@@ -435,12 +441,13 @@ int main(int argc, char *argv[])
 	{
 	    const char *msg = "can not get terminal flags of stdin\n";
 	    int flags;
+	    int ret __attribute__ ((unused));
 
 	    if ((flags = fcntl(0, F_GETFL)) < 0)
 		list_for_each_entry(c, &cons->node, node) {
 		    if (c->fd < 0)
 			continue;
-		    (void)write(c->fd, msg, strlen(msg));
+		    ret = write(c->fd, msg, strlen(msg));
 		}
 	    else {
 		flags &= ~(O_NONBLOCK);
@@ -449,7 +456,7 @@ int main(int argc, char *argv[])
 		    list_for_each_entry(c, &cons->node, node) {
 			if (c->fd < 0)
 			    continue;
-			(void)write(c->fd, msg, strlen(msg));
+			ret = write(c->fd, msg, strlen(msg));
 		    }
 	    }
 	}
@@ -468,11 +475,12 @@ int main(int argc, char *argv[])
 	list_for_each_entry(c, &cons->node, node) {
 	    const char *msg = "blogd: can not fork to become daemon: ";
 	    const char *err = strerror(errno);
+	    int ret __attribute__ ((unused));
 	    if (c->fd < 0)
 		continue;
-	    (void)write(c->fd, msg, strlen(msg));
-	    (void)write(c->fd, err, strlen(err));
-	    (void)write(c->fd, "\n", 1);
+	    ret = write(c->fd, msg, strlen(msg));
+	    ret = write(c->fd, err, strlen(err));
+	    ret = write(c->fd, "\n", 1);
 	}
 	exit(EXIT_FAILURE);
     default:
@@ -492,8 +500,8 @@ int main(int argc, char *argv[])
 	fflush(stdout);
 
 	if (pipefd[0] > 0) {
-	    int dummy;
-	    read(pipefd[0], &dummy, 1);
+	    int dummy, ret __attribute__ ((unused));
+	    ret = read(pipefd[0], &dummy, 1);
 	    close(pipefd[0]);
 	}
 
