@@ -16,6 +16,7 @@
 #include <limits.h>
 #include <linux/magic.h>
 #include <linux/major.h>
+#include <pty.h>
 #include <signal.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -27,6 +28,7 @@
 #include <sys/socket.h>
 #include <sys/stat.h>
 #include <sys/types.h>
+#include <sys/sysmacros.h>
 #include <sys/vfs.h>
 #include <sys/wait.h>
 #include <unistd.h>
@@ -48,6 +50,17 @@
 #endif
 
 int final = 0;
+static volatile char *_arg0;
+
+/*
+ * Avoid trouble if linked with e.g. blogger as there
+ * is no external arg0 but linker on ppc64 and s390/x
+ * seems to expect this.
+ */
+void remember_arg0(volatile char *arg0)
+{
+    _arg0 = arg0;
+}
 
 /*
  * Used to ignore some signals during epoll_pwait(2) or ppoll(2)
@@ -59,9 +72,6 @@ sigset_t omask;
  */
 int epfd  = -1;
 int evmax;
-static int requests = 0;
-
-#define REQUEST_CLOSE	(1<<0)
 
 /*
  * Remember if we're signaled.
@@ -450,9 +460,6 @@ skip:
 	    stop_logging();
 	    flog = close_logging();
 	}
-	if (requests & REQUEST_CLOSE) {
-	    ;
-	}
 	if (nsigio < 0) {
 	    nsigio = SIGIO;
 	    (void)set_signal(SIGIO, NULL, SIG_IGN);
@@ -500,10 +507,6 @@ void closeIO(void)
     stop_logging();
     flog = close_logging();
 
-    if (requests & REQUEST_CLOSE) {
-	;
-    }
-
     if (fdfifo >= 0) {
 	epoll_delete(fdfifo);
 	close(fdfifo);
@@ -531,41 +534,14 @@ void closeIO(void)
     return;
 }
 
-/* Allocate a console */
-static list_t lcons = { &(lcons), &(lcons) };
-static int consalloc(struct console **cons, char *name, const int cflags, const dev_t dev, int io)
+static int consinitIO(struct console *newc)
 {
-    struct console *newc;
-    list_t *head;
     int tflags;
 
-    if (!cons)
-	error("missing console pointer");
-
-    if (posix_memalign((void**)&newc, sizeof(void*), alignof(struct console)+strlen(name)+1) != 0 || !newc)
-	error("memory allocation");
-
-    newc->tty = ((char*)newc)+alignof(struct console);
-    strcpy(newc->tty, name);
-    newc->flags = cflags;
-    newc->dev = dev;
-    newc->pid = -1;
-
-    if (!*cons) {
-	head = &lcons;
-	*cons = (struct console*)head;
-    } else
-	head = &(*cons)->node;
-    insert(&newc->node, head);
-
-    if (!io)
-	return 0;
-
     if ((newc->fd = open(newc->tty, O_WRONLY|O_NONBLOCK|O_NOCTTY)) < 0) {
 	if (errno == EACCES)
 	    error("can not open %s", newc->tty);
 	warn("can not open %s", newc->tty);
-	free(newc);
 	return 0;
     }
 
@@ -585,6 +561,40 @@ static int consalloc(struct console **cons, char *name, const int cflags, const
     return 1;
 }
 
+/* Allocate a console */
+static list_t lcons = { &(lcons), &(lcons) };
+static int consalloc(struct console **cons, char *name, const int cflags, const dev_t dev, int io)
+{
+    struct console *newc;
+    list_t *head;
+
+    if (!cons)
+	error("missing console pointer");
+
+    if (posix_memalign((void**)&newc, sizeof(void*), alignof(struct console)+strlen(name)+1) != 0 || !newc)
+	error("memory allocation");
+
+    newc->tty = ((char*)newc)+alignof(struct console);
+    strcpy(newc->tty, name);
+    newc->flags = cflags;
+    newc->dev = dev;
+    newc->pid = -1;
+
+    if (io && !consinitIO(newc)) {
+	free(newc);
+	return 0;
+    }
+
+    if (!*cons) {
+	head = &lcons;
+	*cons = (struct console*)head;
+    } else
+	head = &(*cons)->node;
+    insert(&newc->node, head);
+
+    return 1;
+}
+
 void getconsoles(struct console **cons, int io)
 {
     static const struct {
@@ -602,6 +612,7 @@ void getconsoles(struct console **cons, int io)
     char fbuf[16], dev[64];
     char *tty = NULL;
     FILE *fc;
+    int items;
 
     if (!cons)
 	error("error: console pointer empty");
@@ -614,11 +625,16 @@ void getconsoles(struct console **cons, int io)
 	goto err;
     }
 
-    while ((fscanf(fc, "%*s %*s (%[^)]) %[0-9:]", &fbuf[0], &dev[0]) == 2)) {
+    while ((items = fscanf(fc, "%*s %*s (%[^)]) %[0-9:]", &fbuf[0], &dev[0]))
+	   != EOF) {
 	char *tmp;
 	int flags, n, maj, min;
 	int ret;
 
+	/* Ignore consoles without tty binding. */
+	if (items != 2)
+	    continue;
+
 	if (!strchr(fbuf, 'E'))
 	    continue;
 
@@ -662,8 +678,10 @@ err:
     if (!tty)
 	error("can not allocate string");
 
-    if (consalloc(&c, tty, CON_CONSDEV, makedev(TTYAUX_MAJOR, 1), io))
-	*cons = c;
+    if (!consalloc(&c, tty, CON_CONSDEV, makedev(TTYAUX_MAJOR, 1), io))
+	error("/dev/console is not a valid fallback\n");
+
+    *cons = c;
 }
 
 /*
@@ -671,7 +689,7 @@ err:
  */
 static void epoll_console_in(int fd)
 {
-    const ssize_t cnt = safein(fdread, trans, sizeof(trans));
+    const ssize_t cnt = safein(fd, trans, sizeof(trans));
     static struct winsize owz;
     struct winsize wz;
 
@@ -923,6 +941,103 @@ static void socket_handler(int fd)
 
 	break;
 
+    case MAGIC_DEACTIVATE:
+
+	if (fdread >= 0) {
+	    struct console *c;
+
+	    list_for_each_entry(c, &cons->node, node) {
+		if (c->flags & CON_CONSDEV) {
+		    if (c->fd > 0) {
+			epoll_delete(fdread);
+			(void)ioctl(c->fd, TIOCCONS, NULL);
+			close(fdread);
+			dup2(c->fd, 0);
+			dup2(0, 1);
+			dup2(0, 2);
+			fdread = -1;
+		    }
+		    break;
+		}
+	    }
+	}
+
+	enqry = ANSWER_ACK;
+	safeout(fd, enqry, strlen(enqry)+1, SSIZE_MAX);
+
+	break;
+
+    case MAGIC_REACTIVATE:
+
+	if (fdread < 0) {
+	    struct console *c;
+
+	    list_for_each_entry(c, &cons->node, node) {
+		if (c->flags & CON_CONSDEV) {
+		    if (c->fd > 0) {
+			char ptsname[NAME_MAX+1];
+			struct termios o;
+			struct winsize w;
+			struct stat st;
+			speed_t ospeed = B38400;
+			speed_t ispeed = B38400;
+			int ptm, pts;
+
+			w.ws_row = 0;
+			w.ws_col = 0;
+			if (ioctl(c->fd, TIOCGWINSZ, &w) < 0)
+			    error("can not get window size of %s", c->tty);
+
+			if (!w.ws_row)
+			    w.ws_row = 24;
+			if (!w.ws_col)
+			    w.ws_col = 80;
+
+			memcpy(&o, &c->otio, sizeof(o));
+			cfmakeraw(&o);
+			cfsetispeed(&o, ispeed);
+			cfsetospeed(&o, ospeed);
+			o.c_lflag &= ~ECHO;
+			o.c_lflag |= ISIG;
+			o.c_cc[VTIME] = 0;
+			o.c_cc[VMIN]  = CMIN;
+
+			if (openpty(&ptm, &pts, ptsname, &o, &w) < 0)
+			    error("can not open pty/tty pair");
+
+			if (fstat(pts, &st) < 0)
+			    error("can not stat slave pty");
+			else {
+			    struct termios lock;
+			    memset(&lock, 0xff, sizeof(lock));
+			    (void)ioctl(pts, TIOCSLCKTRMIOS, &lock);
+			}
+
+			if (ioctl(pts, TIOCCONS, NULL) < 0)
+			    error("can not set console device to %s", ptsname);
+
+			dup2(ptm, 0);
+			if (ptm > 0)
+			    close(ptm);
+
+			dup2(pts, 1);
+			dup2(pts, 2);
+			if (pts > 2)
+			    close(pts);
+
+			epoll_addread(0, &epoll_console_in);
+			fdread = 0;
+		    }
+		    break;
+		}
+	    }
+	}
+
+	enqry = ANSWER_ACK;
+	safeout(fd, enqry, strlen(enqry)+1, SSIZE_MAX);
+
+	break;
+
     case MAGIC_QUIT:
 
 	enqry = ANSWER_ACK;
@@ -934,9 +1049,31 @@ static void socket_handler(int fd)
 
 	break;
 
-    case MAGIC_CLOSE:
+    case MAGIC_FINAL:
 
-	requests |= REQUEST_CLOSE;
+	enqry = ANSWER_ACK;
+	safeout(fd, enqry, strlen(enqry)+1, SSIZE_MAX);
+
+	if (!final && _arg0) {
+	    int ret;
+
+	    final = 1;
+
+	    if (_arg0[0] != '@')
+		_arg0[0] = '@';
+
+	    ret = rename(BOOT_LOGFILE, BOOT_OLDLOGFILE);
+	    if (ret < 0) {
+		if (errno == EACCES || errno == EROFS || errno == EPERM)
+		    goto skip;
+		if (errno != ENOENT)
+		    error("Can not rename %s", BOOT_LOGFILE);
+	    }
+	}
+    skip:
+	break;
+
+    case MAGIC_CLOSE:
 
 	enqry = ANSWER_ACK;
 	safeout(fd, enqry, strlen(enqry)+1, SSIZE_MAX);
